# Public: Exposes a simple read-eval-print loop interface.
module Pal
  # Public: The current version of Pal.
  Version = "0.1.0"

  # Public: Defines the evaluation context. All instance methods defined in
  # this class are made available as top-level methods during evaluation.
  class Context
    # Internal: Stores the result of the last-evaluated expression.
    attr_reader :_

    # Internal: Provides a less verbose string representation for `Context`s.
    def inspect
      "Context"
    end
  end

  # Public: Creates a REPL instance.
  class REPL
    # Public: Gets the name of the REPL instance, used to generate the prompt
    # text and debugging information.
    attr_reader :name

    # Public: Gets the evaluation context associated with the REPL instance.
    attr_reader :context

    # Public: Gets the result of the last-evaluated expression.
    attr_reader :result

    def initialize(name, context)
      @name = name
      @context = context
      @line = 1
      @statements = []
    end

    # Public: Starts the REPL. The REPL will persist until an `exit` signal
    # is received.
    def loop
      catch :exit do
        read while true
      end
      self
    end

    # Public: Gets a line from the standard input using Readline.
    def get(prompt)
      require "readline"
      Readline.readline("#{prompt}> ", true)
    end

    # Public: Writes a line to the standard output.
    alias_method :put, :puts

    # Public: Reads, evaluates, and prints a line of Ruby code. If the line is
    # `exit` or `quit`, an `exit` signal is sent and the loop terminates. If an
    # interrupt is sent via Control-C, a blank line is printed before exiting.
    # Multiple lines of code may be evaluated by terminating each line with an
    # escape character (`\`).
    def read(prompt = @name)
      @error = nil
      @input = get prompt
      throw :exit if @input.nil? || %w{exit quit}.include?(@input)
      unless @input.empty?
        @statements << @input
        if @statements.last[-1] == "\\"
          @statements.last.chop!
          read "." * @name.size
        else
          @input = @statements.join($/)
          @statements.clear
          evaluate @input
          print @result
        end
      end
    rescue Interrupt
      puts
    end

    # Public: Evaluates a line of code in the current context. If an error was
    # raised, the stack trace is printed. To reduce noise, all lines containing
    # the current file name are omitted from the trace.
    def evaluate(value)
      @result = eval(value, @context.instance_eval { binding }, "(#{@name})", @line)
      @context.instance_variable_set(:@_, result)
    rescue Exception => exception
      @error = true
      trace = exception.backtrace.reject { |line| line.include? File.basename(__FILE__) }.unshift(exception.message)
      warn "#{exception.class}: #{trace.join("\n\t")}"
    ensure
      @line += 1
    end

    # Public: Prints the last output, prepending the output symbol to a
    # human-readable representation of the result.
    def print(result)
      put "=> #{result.inspect}" unless @error
    end
  end
end